
import numpy as np
import adata_query


class SelfIDMixIn(object):
    @property
    def _NN_0_idx(self) -> np.ndarray:
        """
        Return the idx of the first neighbor cell identified from the query, for
        each cell in the query.
        """
        return self._X_nn[:, 0].flatten()

    @property
    def _X_NN_0(self) -> np.ndarray:
        """
        Fetch the first neighbor cell from query, for each query cell.
        """
        return adata_query.fetch(
            self._adata[self._NN_0_idx],
            key=self._use_key,
            torch=False,
        )

    @property
    def _self_id(self) -> bool:
        """
        Answer whether all of the first cells identified from the nn query
        is that same as the input query.
        """
        return np.all(self._X_NN_0 == self._X_query)

    def check_self_id(
        self, X_query: np.ndarray, X_nn: np.ndarray, *args, **kwargs
    ) -> bool:
        
        """
        Parameters:
        -----------
        X_query: np.ndarray
        
        X_nn: np.ndarray
        
        Returns
        -------
        self_id: bool
        """
        
        self.__update__(locals())

        return self._self_id



import ABCParse
import adata_query
import anndata
import annoy
import numpy as np
import time

from ._type_check_mix_in import TypeCheckMixIn
from . import __logger__


from typing import Union
        
class BasekNN(TypeCheckMixIn, ABCParse.ABCParse):
    _IDX_BUILT = False

    def __init__(self, *args, **kwargs):
        
        super().__init__()
        
        __logger__.info(f"{self.__class__.__name__}.__init__")


    @property
    def _X_basis(self) -> np.ndarray:
        """Basis data on which the graph is built."""
        if not hasattr(self, "_X_basis_ATTR"):
            self._X_basis_ATTR = adata_query.fetch(
                self._adata, key=self._use_key, torch=False
            )
        return self._X_basis_ATTR

    @property
    def _N_CELLS(self) -> int:
        return self._X_basis.shape[0]

    @property
    def _N_DIMS(self) -> int:
        return self._X_basis.shape[1]

    def _build_annoy_index(self) -> annoy.AnnoyIndex:
        
        __logger__.info("Start kNN build")
        
        self._start_build_time = time.time()

        idx = annoy.AnnoyIndex(self._N_DIMS, self._metric)
        [idx.add_item(i, self._X_basis[i]) for i in range(self._N_CELLS)]
        idx.build(self._n_trees)
        self._idx = idx
        self._IDX_BUILT = True

        self._end_build_time = time.time()
        
        __logger__.info("End kNN build")

    @property
    def _BUILD_TIME(self) -> str:
        if not hasattr(self, "_build_time"):
            self._build_time = self._end_build_time - self._start_build_time
            self._print_build_time = "{:.4f} s".format(self._build_time)
            __logger__.info(f"kNN build time: {self._print_build_time}")
        return self._print_build_time

    @property
    def _BUILD_TIME_PER_CELL(self) -> str:
        self._print_build_time_per_cell = "{:.2e} s".format(self._build_time / self._N_CELLS)
        __logger__.info(f"kNN build time per cell: {self._print_build_time_per_cell}")
        return self._print_build_time_per_cell

    @property
    def idx(self) -> annoy.AnnoyIndex:
        """The actual annoy.AnnoyIndex"""

        if not hasattr(self, "_idx"):
            self._build_annoy_index()
        return self._idx

    def __repr__(self):
        return f"kNN: {self._use_key}, {self._N_CELLS} cells"



import inspect

from typing import Any, Dict

from . import __logger__


class TypeCheckMixIn(object):
    _enabled = True
    
    def __init__(self, *args, **kwargs):
        
        """enabled by default"""
        
        __logger__.info(msg=f"{self._cls_name}:TypeChecker ENABLED")
        
    @property
    def _cls_name(self):
        return self.__class__.__name__
    
    def _disable_type_check(self):
        self._enabled = False
        __logger__.info(msg=f"{self._cls_name}:TypeChecker DISABLED.")
        
    def _enable_type_check(self):
        self._enabled = True
        __logger__.info(msg=f"{self._cls_name}:TypeChecker DISABLED.")
    
    @property
    def _TYPECHECK_IGNORE(self):
        return ["self", "args", "kwargs"]

    @property
    def _SIGNATURE(self):
        return inspect.signature(self._func).parameters.items()

    def __check_types__(self, func: Any, passed: Dict):
        """
        func
        passed
        """

        if self._enabled:
            self._func = func

            for key, val in self._SIGNATURE:
                if not key in self._TYPECHECK_IGNORE:
                    assert isinstance(passed[key], val.annotation), TypeError(
                        f"arg: {key} is not of correct type: {val}"
                    )




import ABCParse
import pandas as pd
import numpy as np

from typing import Union


class ValueCountAssignment(ABCParse.ABCParse):
    def __init__(self, *args, **kwargs):
        self.__parse__(locals())

    def forward(self):
        return (
            pd.DataFrame(self._neighbor_vals.squeeze(-1))
            .apply(pd.Series.value_counts, axis=1)
            .fillna(0)
        )

    def max_assign(self, value_counts):
        return value_counts.idxmax(1).values

    def __call__(
        self, neighbor_vals: np.ndarray, max_only: bool = True, *args, **kwargs
    ):
        self.__update__(locals())

        value_counts = self.forward()
        if max_only:
            return self.max_assign(value_counts)
        return value_counts


def count_based_assignment(
    neighbor_vals: np.ndarray, max_only: bool = True, *args, **kwargs
) -> Union[np.ndarray, pd.DataFrame]:
    """

    Parameters
    ----------
    neighbor_vals: np.ndarray

    max_only: bool, default = True

    Returns
    -------
    consensus_assignment: np.ndarray or value_counts: pd.DataFrame
    """
    value_count_assignment = ValueCountAssignment()
    return value_count_assignment(neighbor_vals=neighbor_vals, max_only=max_only)

class CountBasedAssignmentMixIn(object):
    def count_based_assignment(self, neighbor_vals: np.ndarray, max_only: bool = True):
        return count_based_assignment(neighbor_vals=neighbor_vals, max_only=max_only)




# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import annoy
import numpy as np


# -- import local dependencies: -----------------------------------------------
from ._type_check_mix_in import TypeCheckMixIn
from . import __logger__
from ._base_knn import BasekNN

from typing import List, Union


# -- operating class: ---------------------------------------------------------
class NeighborQuery(TypeCheckMixIn, ABCParse.ABCParse):
    def __init__(
        self,
        idx: annoy.AnnoyIndex,
        k_neighbors: int = 20,
        type_check_enabled: bool = True,
        *args,
        **kwargs
    ) -> None:
        
        """
        Parameters
        ----------
        idx: annoy.AnnoyIndex
        
        k_neighbors: int, default = 20
        
        type_check_enabled: bool, default = True
        
        Returns
        -------
        None
        """
        
        super().__init__()
        
        __logger__.info(f"{self.__class__.__name__}.__init__")
        
        self.__parse__(locals())
        self.__check_types__(func=self.__init__, passed=locals())
        
        if not type_check_enabled:
            self._disable_type_check()
            
    @property
    def k(self) -> int:
        return int(self._k_neighbors + self._offset)

    @property
    def _N_QUERY_CELLS(self) -> int:
        return self._X_query.shape[0]

    def _forward(self, X_query_i: np.ndarray) -> List:
        return self._idx.get_nns_by_vector(X_query_i, self.k)[self._offset:]
    
    def forward(self, X_query: np.ndarray) -> np.ndarray:
        return np.array([self._forward(X_query[i]) for i in range(self._N_QUERY_CELLS)])

    def __call__(self, X_query: np.ndarray, offset: int = 0) -> np.ndarray:
        """
        Match observed cells (X_query: np.ndarray) to their (k) nearest
        neighbors (X_nn: np.ndarray) in the given annoy idx.
        
        Parameters
        ----------
        X_query: np.ndarray
            Observations of size: [cells x n_dims].
            
        offset: int, default = 0
            Adds a pseudocount. Particularly useful for removing self-identifying cells.

        Returns
        -------
        X_nn: np.ndarray
            Neighbor assignments of size: [cells x n_neighbors].
        """

        __logger__.info(f"{self.__class__.__name__}.__call__")
        
        self.__update__(locals())
        self.__check_types__(func=self.__call__, passed=locals())
        
        return self.forward(X_query)


def _index_flatten(kNN: Union[annoy.AnnoyIndex, BasekNN]) -> annoy.AnnoyIndex:
    
    """Enable use of either cell_ann.kNN or annoy.AnnoyIndex
    
    Returns
    -------
    idx: annoy.AnnoyIndex
    """
    
    if isinstance(kNN, annoy.AnnoyIndex):
        return kNN
    elif isinstance(kNN, BasekNN):
        return kNN.idx
    
def neighbor_query(
    kNN: Union[annoy.AnnoyIndex, BasekNN],
    X_query: np.ndarray,
    k_neighbors: int = 20,
    offset: int = 0,
    type_check_enabled: bool = True,
) -> np.ndarray:
    """
    Parameters
    ----------
    kNN: Union[annoy.AnnoyIndex, BasekNN]
    
    X_query: np.ndarray
    
    k_neighbors: int = 20
    
    offset: int = 0
    
    type_check_enabled: bool = True
    
    
    Returns
    -------
    X_nn: np.ndarray
    """
    
    idx = _index_flatten(kNN)
        
    knn_query = NeighborQuery(
        idx=idx,
        k_neighbors=k_neighbors,
        type_check_enabled=type_check_enabled,
    )
    return knn_query(X_query=X_query, offset = offset)

class NeighborQueryMixIn(object):
    
    def neighbor_query(
        self: Union[annoy.AnnoyIndex, BasekNN],
        X_query: np.ndarray,
        k_neighbors: int = 20,
        offset: int = 0,
        type_check_enabled: bool = True,
    ) -> np.ndarray:
        """
        
        """
        return neighbor_query(
            kNN=self,
            X_query=X_query,
            k_neighbors=k_neighbors,
            offset=offset,
            type_check_enabled=type_check_enabled,
        )



# __init__.py

from ._logger import Logger as _LOGGER
from .__version__ import __version__

__logger__ = _LOGGER()

import pathlib

pkg_path = pathlib.Path(__file__).absolute()
__logger__.info(f"import cell-ann (version: {__version__})")
__logger__.info(f"source: {pkg_path}")

from ._neighbor_query import NeighborQuery, neighbor_query
from ._base_knn import BasekNN
from ._knn import kNN
from ._query_mapper import map_query
from ._value_counts_assignment import count_based_assignment



import ABCParse
import anndata
import adata_query
import numpy as np
import pandas as pd

from typing import Dict, List, Union


class QueryMapper(ABCParse.ABCParse):
    """
    # NeighborRetriever
    Map retrieved neighbors to values in reference adata.

    With the neighbor indices in hand, one can retrieve
    reference adata attributes
    """

    def __init__(self, adata: anndata.AnnData, *args, **kwargs):
        self.__parse__(locals())

    @property
    def _FETCH_KWARGS(self):
        kwargs = self._PARAMS.copy()
        params_to_pop = ["args", "adata"]
        [kwargs.pop(param) for param in params_to_pop]
        return ABCParse.function_kwargs(func=adata_query.fetch, kwargs=kwargs)

    @property
    def flat_nn_idx(self):
        return self._X_nn.flatten()

    @property
    def _N(self):
        return len(self.flat_nn_idx)

    @property
    def _N_CELLS(self):
        return self._X_nn.shape[0]

    @property
    def _N_NEIGHBORS(self):
        return self._X_nn.shape[1]

    @property
    def nn_adata(self):
        """Setting this as an attribute enables fast recall of multiple items
        (e.g.; something from obs and "X_pca")"""
        if not hasattr(self, "_nn_adata"):
            self._nn_adata = self._adata[self.flat_nn_idx]
        return self._nn_adata

    def forward(self, key):
        X_fetch = adata_query.fetch(adata=self.nn_adata, key=key, **self._FETCH_KWARGS)
        if isinstance(X_fetch, pd.Series):
            X_fetch = X_fetch.to_numpy()
        return X_fetch.reshape(self._N_CELLS, self._N_NEIGHBORS, -1)  # )

    def __call__(
        self,
        X_nn: np.ndarray,
        attr_keys: Union[List[str], str],
        torch: bool = False,
        *args,
        **kwargs
    ) -> np.ndarray:
        """

        This part is meant to be flexible. Others can be more rigid.

        Parameters
        ----------
        X_nn: np.ndarray

        attr_keys: Union[List[str], str],

        torch: bool = False

        Returns
        -------
        """
        self.__update__(locals())

        return {key: self.forward(key) for key in ABCParse.as_list(attr_keys)}


def map_query(
    adata: anndata.AnnData, X_nn, attr_keys: Union[List[str], str]
) -> Dict[str, np.ndarray]:
    """
    This part is meant to be flexible (i.e., always returning a dictionary of
    np.ndarray of shape: n_cell x n_neighbors x n_dim_feature) Others can be more rigid.
    """
    qmap = QueryMapper(adata)
    return qmap(X_nn, attr_keys)


class MapQueryMixIn(object):
    def map_query(
        self, X_nn, attr_keys: Union[List[str], str]
    ) -> Dict[str, np.ndarray]:
        return map_query(adata=self._adata, X_nn=X_nn, attr_keys=attr_keys)



import adata_query
import anndata
from . import __logger__
import numpy as np

from ._base_knn import BasekNN
from ._self_id_mix_in import SelfIDMixIn
from ._neighbor_query import NeighborQueryMixIn
from ._query_mapper import MapQueryMixIn
from ._value_counts_assignment import CountBasedAssignmentMixIn

class kNN(
    BasekNN,
    SelfIDMixIn,
    NeighborQueryMixIn,
    MapQueryMixIn,
    CountBasedAssignmentMixIn,
):
    def __init__(
        self,
        adata: anndata.AnnData,
        use_key: str = "X_pca",
        metric: str = "euclidean",
        n_neighbors: int = 20,
        n_trees: int = 10,
        type_check_enabled: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__()

        __logger__.info(f"{self.__class__.__name__}.__init__")

        self.__parse__(locals())

        if not type_check_enabled:
            self._disable_type_check()

        self.__check_types__(func=self.__init__, passed=locals())
        self._build_annoy_index()
    
    def map_obs(
        self,
        X_query: np.ndarray,
        key: str,
        k_neighbors: int = 20,
        offset: int = 0,
        type_check_enabled: bool = True,
    ) -> np.ndarray:
        
        X_nn = self.neighbor_query(
            X_query=X_query, 
            k_neighbors=k_neighbors,
            offset=offset,
            type_check_enabled=type_check_enabled,
        )
                              
        return self.count_based_assignment(
            self.map_query(X_nn, attr_keys=key)[key], max_only=True
        )



import ABCParse
import logging
import pathlib
import os

from typing import Union, Optional


class Logger(ABCParse.ABCParse):
    def __init__(
        self,
        name: Optional[Union[str, None]] = "cell-ann-logger",
        wd: Union[pathlib.Path, str] = pathlib.Path(os.getcwd()),
        log_dir: Optional[Union[pathlib.Path, str]] = None,
        level: int = logging.DEBUG,
        format: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        *args,
        **kwargs
    ):
        """"""

        self.__parse__(locals())
        self._configure_logging()

    @property
    def _WD(self):
        if not isinstance(self._wd, pathlib.Path):
            self._wd = pathlib.Path(self._wd)
        return self._wd
    
    @property
    def _LOG_DIR(self):
        if not hasattr(self, "_log_dir") or (self._log_dir is None):
            self._log_dir = self._WD.joinpath("logs")
        if not self._log_dir.exists():
            self._log_dir.mkdir()
        return self._log_dir

    @property
    def _BASIC_LOG_PATH(self):
        return self._LOG_DIR.joinpath("cell_ann.log")
    
    @property
    def _EXISTING_LOG_PATHS(self):
        return list(self._BASIC_LOG_PATH.parent.glob("cell_ann*log"))
    
    @property
    def _N_LOGS(self):
        return len(self._EXISTING_LOG_PATHS)
    
    @property
    def _LOG_PATH(self):
        return self._BASIC_LOG_PATH.parent.joinpath(f"cell_ann.{self._N_LOGS}.log")

    def _configure_logging(self):
        logging.basicConfig(
            filename=self._LOG_PATH,
            filemode="a",
            level=self._level,
            format=self._format,
        )
        # disable logging from nb_black
        for logger_key in ['h5py', 'blib2to3']:
            logging.getLogger(logger_key).setLevel(logging.ERROR)

    @property
    def logger(self):
        if not hasattr(self, "_logger"):
            self._logger = logging.getLogger(self._name)
        return self._logger

    def debug(self, msg):
        self.logger.debug(msg)

    def info(self, msg):
        self.logger.info(msg)

    def warning(self, msg):
        self.logger.warning(msg)

    def error(self, msg):
        self.logger.error(msg)

    def critical(self, msg):
        self.logger.critical(msg)
